/*
  Copyright Red Hat, Inc. 2005-2008

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2, or (at your option) any
  later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; see the file COPYING.  If not, write to the
  Free Software Foundation, Inc.,  675 Mass Ave, Cambridge,
  MA 02139, USA.
*/
/*
 * Author: Stanko Kupcevic <kupcevic@redhat.com>
 */


#include "SSLClient.h"
#include "Mutex.h"
#include "Time.h"
#include "Random.h"
#include "utils.h"
#include "File.h"

#include <errno.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>
#include <unistd.h>

#include <iostream>
#include <list>
#include <set>



#include <openssl/err.h>



using namespace std;



static Mutex global_lock;
static bool ssl_inited = false;
static SSL_CTX* ctx = 0;
static vector<counting_auto_ptr<Mutex> > ssl_locks;

class file_cert
{
public:
  file_cert(const String& file, const String& cert) :
    file(file),
    cert(cert) {}

  String file;
  String cert;
};
static list<file_cert> trusted_certs;



static int
verify_cert_callback(int preverify_ok, X509_STORE_CTX *ctx)
{
  return 1;
}
static void
load_peer_certs()
{
  MutexLocker l(global_lock);

  // load trusted CAs
  if (!SSL_CTX_load_verify_locations(ctx,
				     _trust_CAs,
				     NULL))
    ;//cout << "failed to load trusted CAs" << endl;

  STACK_OF(X509_NAME) *cert_names =
    SSL_load_client_CA_file(_trust_CAs);
  if (cert_names)
    SSL_CTX_set_client_CA_list(ctx, cert_names);
//  else
//    cout << "failed to load trusted CAs" << endl;

  // load saved certs

  set<String> files;
  String dir_path(_certs_store_dir);
  DIR* d = opendir(dir_path.c_str());
  if (d == NULL)
    throw String("unable to open directory ") + dir_path;
  try {
    while (true) {
      struct dirent* ent = readdir(d);
      if (ent == NULL) {
	closedir(d);
	break;
      }
      String kid_path = ent->d_name;
      if (kid_path == "." || kid_path == "..")
	continue;
      kid_path = dir_path + "/" + kid_path;
      struct stat st;
      if (stat(kid_path.c_str(), &st))
	continue;
      if (S_ISREG(st.st_mode))
	files.insert(kid_path);
    }
  } catch ( ... ) {
    closedir(d);
    throw;
  }

  trusted_certs.clear();

  for (set<String>::const_iterator iter = files.begin();
       iter != files.end();
       iter++) {
    try {
      String cert(File::open(*iter).read());
      if (cert.size() && cert.size() < 10 * 1024)
	trusted_certs.push_back(file_cert(*iter, cert));
    } catch ( ... ) {}
  }
}
static void
ssl_mutex_callback(int mode,
		   int n,
		   const char *file,
		   int line)
{
  if (mode & CRYPTO_LOCK)
    ssl_locks[n]->lock();
  else
    ssl_locks[n]->unlock();
}
static pthread_t
ssl_id_callback(void)
{
  return pthread_self();
}




// ##### class SSLClient #####


SSLClient::SSLClient(ClientSocket sock) :
  _sock(sock),
  _connected(false)
{
  {
    MutexLocker l(global_lock);
    if (!ssl_inited) {
      // init library

      SSL_library_init();
      // TODO: random number generator,
      // not on systems with /dev/urandom (eg. Linux)

      // thread support
      ssl_locks.clear();
      for (int i=0; i<CRYPTO_num_locks()+1; i++)
	ssl_locks.push_back(counting_auto_ptr<Mutex>(new Mutex()));
      CRYPTO_set_locking_callback(ssl_mutex_callback);
      CRYPTO_set_id_callback(ssl_id_callback);

      // create context
      if (!ctx)
	ctx = SSL_CTX_new(SSLv23_client_method());
      if (!ctx)
	throw String("SSL context creation failed");
      // set verify_callback() function
      SSL_CTX_set_verify(ctx,
			 SSL_VERIFY_PEER,
			 verify_cert_callback);
      // set mode
      SSL_CTX_set_mode(ctx, SSL_MODE_ENABLE_PARTIAL_WRITE);
      SSL_CTX_set_mode(ctx, SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER);

      // load key
      if (!SSL_CTX_use_PrivateKey_file(ctx,
				       _privkey,
				       SSL_FILETYPE_PEM))
	throw String("error importing cert key file");
      // load cert
      if (!SSL_CTX_use_certificate_file(ctx,
					_pubkey,
					SSL_FILETYPE_PEM))
	throw String("error importing cert file");
      // load peers' certs
      load_peer_certs();

      ssl_inited = true;
    }

    // create SSL object, giving it context
    _ssl = SSL_new(ctx);
    if (!_ssl)
      throw String("creation of ssl object failed");
  }

  // make socket non-blocking
  try {
    _sock.nonblocking(true);
  } catch ( ... ) {
    SSL_free(_ssl);
    throw;
  }

  // assign fd to _ssl
  if (!SSL_set_fd(_ssl, _sock.get_sock())) {
    SSL_free(_ssl);
    throw String("fd assignment to ssl_obj failed");
  }
}

SSLClient::~SSLClient()
{
  SSL_shutdown(_ssl);
  SSL_free(_ssl);
}


bool
SSLClient::connect(unsigned int timeout)
{
  if (_connected)
    return _connected;

  unsigned int beg = time_mil();
  while (time_mil() < beg + timeout) {
    int ret = SSL_connect(_ssl);
    if (ret == 1) {
      _connected = true;
      break;
    } else {
      bool want_read, want_write;
      check_error(ret, want_read, want_write);
      socket().ready(want_read, want_write, 250);
    }
  }

  return _connected;
}

String
SSLClient::send(const String& msg,
		unsigned int timeout)
{
  if (!_connected)
    throw String("cannot send, yet: SSL connection not connected");

  if (msg.empty())
    return msg;

  unsigned int beg = time_mil();
  while (time_mil() < beg + timeout) {
    int ret = SSL_write(_ssl, msg.c_str(), msg.size());
    if (ret > 0) {
      return msg.substr(ret);
    } else {
      bool want_read, want_write;
      check_error(ret, want_read, want_write);
      socket().ready(want_read, want_write, 250);
    }
  }

  return msg;
}

String
SSLClient::recv(unsigned int timeout, size_t& buflen)
{
  if (!_connected)
    throw String("cannot receive, yet: SSL connection not connected");

  char buff[4096];

  buflen = sizeof(buff);
  unsigned int beg = time_mil();
  while (time_mil() < beg + timeout) {
    int ret = SSL_read(_ssl, buff, sizeof(buff));
    if (ret > 0) {
      String data(buff, ret);
      memset(buff, 0, sizeof(buff));
      return data;
    } else {
      bool want_read, want_write;
      check_error(ret, want_read, want_write);
      socket().ready(want_read, want_write, 250);
    }
  }

  return "";
}

bool
SSLClient::peer_has_cert()
{
  if (!_connected)
    throw String("cannot determine if peer has certificate: SSL connection not connected");

  if (_cert_pem.size())
    return true;

  X509* cert = SSL_get_peer_certificate(_ssl);
  if (!cert)
    return false;

  // load cert into _cert_pem
  FILE* f = NULL;
  try {
    if (!(f = tmpfile()))
      throw String("unable to open temp file");
    if (!PEM_write_X509(f, cert))
      throw String("unable to write cert to tmp file");
    X509_free(cert); cert = NULL;

    // read cert
    rewind(f);
    while (true) {
      char buff[1024];
      size_t i = fread(buff, sizeof(char), sizeof(buff), f);
      _cert_pem.append(buff, i);
      if (i == 0) {
	if (feof(f))
	  break;
	else
	  throw String("error while reading certificate from temp file");
      }
    }
    fclose(f); f = NULL;
  } catch ( ... ) {
    if (cert)
      X509_free(cert);
    if (f)
      fclose(f);
    _cert_pem.clear();
    throw;
  }

  return true;
}

String
SSLClient::peer_cert_fingerprint(String& digest)
{
  if (!peer_has_cert())
    throw String("peer did not present cert");

  String f_name("/tmp/luci_tmp_XXXXXX");
  int fd = -1;
  char* buff = new char[f_name.size() + 1];
  try {
    // pick a filename
    strcpy(buff, f_name.c_str());
    if ((fd = mkstemp(buff)) == -1)
      throw String("unable to generate random file");
    f_name = buff;
    delete[] buff; buff = 0;
    while (close(fd) && errno == EINTR) ; fd = -1;

    File f = File::open(f_name, true);
    f.replace(_cert_pem);

    String out, err;
    int status;
    vector<String> args;
    args.push_back("x509");
    args.push_back("-sha1");
    args.push_back("-in");
    args.push_back(f_name);
    args.push_back("-noout");
    args.push_back("-fingerprint");
    if (utils::execute("/usr/bin/openssl",
		       args,
		       out,
		       err,
		       status,
		       false))
      throw command_not_found_error_msg("/usr/bin/openssl");
    if (status)
      throw String("openssl command failed");
    unlink(f_name.c_str());

    vector<String> words(utils::split(utils::strip(out)));
    if (words.size() != 2)
      throw String("error parsing fingerprint");

    String finger(words[1]);
    String::size_type idx = finger.find('=');
    if (idx == finger.npos ||
	idx+1 == finger.size())
      throw String("error parsing fingerprint");

    digest = words[0];
    return finger.substr(idx+1);
  } catch ( ... ) {
    delete[] buff;
    if (fd != -1)
      while (close(fd) && errno == EINTR)
	;
    unlink(f_name.c_str());
    throw;
  }
}

bool
SSLClient::peer_cert_trusted()
{
  // signed by trusted CAs?
  X509* cert = SSL_get_peer_certificate(_ssl);
  if (!cert)
    return false;
  X509_free(cert);
  if (SSL_get_verify_result(_ssl) == X509_V_OK)
    return true;

  // cert present among saved certs?
  peer_has_cert();  // make sure cert is saved in _cert_pem
  MutexLocker l(global_lock);
  for (list<file_cert>::const_iterator iter = trusted_certs.begin();
       iter != trusted_certs.end();
       iter++)
    if (iter->cert == _cert_pem)
      return true;
  return false;
}

bool
SSLClient::trust_peer_cert(const String& hostname)
{
  MutexLocker l(global_lock);

  if (peer_cert_trusted())
    return true;

  if (!peer_has_cert())
    throw String("peer did not present cert");

  String filename(_certs_store_dir);
  filename += "/" + hostname + "_cert_pub";
  try {
    File::create(filename).replace(_cert_pem);
  } catch ( ... ) {
    unlink(filename.c_str());
    return false;
  }

  load_peer_certs();

  return true;
}

bool
SSLClient::untrust_peer_cert()
{
  MutexLocker l(global_lock);

  if (!peer_has_cert())
    throw String("peer did not present cert");

  for (list<file_cert>::const_iterator iter = trusted_certs.begin();
       iter != trusted_certs.end();
       iter++)
    if (iter->cert == _cert_pem)
      unlink(iter->file.c_str());

  load_peer_certs();
  return true;
}

ClientSocket&
SSLClient::socket()
{
  return _sock;
}

void
SSLClient::check_error(int value, bool& want_read, bool& want_write)
{
  want_read = want_write = false;
  int saved_errno = errno;

  String e;
  switch (SSL_get_error(_ssl, value)) {
  case SSL_ERROR_NONE:
    e = "SSL_ERROR_NONE";
    return;
  case SSL_ERROR_ZERO_RETURN:
    e = "SSL_ERROR_ZERO_RETURN";
    break;
  case SSL_ERROR_WANT_READ:
    want_read = true;
    return;
  case SSL_ERROR_WANT_WRITE:
    want_write = true;
    return;
  case SSL_ERROR_WANT_CONNECT:
    e = "SSL_ERROR_WANT_CONNECT";
    break;
  case SSL_ERROR_WANT_ACCEPT:
    e = "SSL_ERROR_WANT_ACCEPT";
    break;
  case SSL_ERROR_WANT_X509_LOOKUP:
    e = "SSL_ERROR_WANT_X509_LOOKUP";
    break;
  case SSL_ERROR_SYSCALL:
    if (saved_errno == EAGAIN || saved_errno == EINTR)
        return;
    e = "SSL_ERROR_SYSCALL";
    throw String("SSL error: ") + e + ":" + String(strerror(saved_errno));
    break;
  case SSL_ERROR_SSL:
	char buf[2048];
    e = "SSL_ERROR_SSL";
	ERR_error_string_n(ERR_get_error(), buf, sizeof(buf));
	throw String("SSL error: ") + e + ":" + String(buf);
    break;
  }

  throw String("SSL error: ") + e;

}
