/*
  Copyright Red Hat, Inc. 2006-2008

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2, or (at your option) any
  later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; see the file COPYING.  If not, write to the
  Free Software Foundation, Inc.,  675 Mass Ave, Cambridge,
  MA 02139, USA.
*/
/*
 * Author: Stanko Kupcevic <kupcevic@redhat.com>
 */


#include <Python.h>

#include <Socket.h>
#include <String.h>
#include <Time.h>
#include <XML.h>
#include <utils.h>

#include "SSLClient.h"

#include <map>

using namespace std;


class PythonThreadsAllower
{
public:
  PythonThreadsAllower()
  { _save = PyEval_SaveThread(); }

  ~PythonThreadsAllower()
  { PyEval_RestoreThread(_save); }

private:
  PyThreadState *_save;
};



static Mutex mutex;
static map<int, counting_auto_ptr<SSLClient> > ssls;



static PyObject *
conga_ssl_lib_connect(PyObject *self, PyObject *args);
static PyObject *
conga_ssl_lib_disconnect(PyObject *self, PyObject *args);

static PyObject *
conga_ssl_lib_send(PyObject *self, PyObject *args);
static PyObject *
conga_ssl_lib_recv(PyObject *self, PyObject *args);

static PyObject *
conga_ssl_lib_trust(PyObject *self, PyObject *args);
static PyObject *
conga_ssl_lib_trusted(PyObject *self, PyObject *args);
static PyObject *
conga_ssl_lib_untrust(PyObject *self, PyObject *args);

static PyObject *
conga_ssl_lib_peer_fingerprint(PyObject *self, PyObject *args);



static PyMethodDef SSLMethods[] = {
  {"connect", conga_ssl_lib_connect, METH_VARARGS,
   "doc"},
  {"disconnect", conga_ssl_lib_disconnect, METH_VARARGS,
   "doc"},
  {"send", conga_ssl_lib_send, METH_VARARGS,
   "doc"},
  {"recv", conga_ssl_lib_recv, METH_VARARGS,
   "doc"},
  {"trust", conga_ssl_lib_trust, METH_VARARGS,
   "doc"},
  {"trusted", conga_ssl_lib_trusted, METH_VARARGS,
   "doc"},
  {"untrust", conga_ssl_lib_untrust, METH_VARARGS,
   "doc"},
  {"peer_fingerprint", conga_ssl_lib_peer_fingerprint, METH_VARARGS,
   "doc"},
  {NULL, NULL, 0, NULL}        /* Sentinel */
};



PyMODINIT_FUNC
initconga_ssl_lib(void)
{
  (void) Py_InitModule("conga_ssl_lib", SSLMethods);
}



PyObject *
conga_ssl_lib_connect(PyObject *self, PyObject *args)
{
  const char* hostname;
  int port;
  int timeout;
  if (!PyArg_ParseTuple(args,
			"sii",
			&hostname,
			&port,
			&timeout))
    return NULL;
  if (port < 1 || port > 65535) {
    PyErr_SetString(PyExc_ValueError, "invalid port number");
    return NULL;
  }
  if (timeout < 0) {
    PyErr_SetString(PyExc_ValueError, "negative timeout");
    return NULL;
  }

  try {
    counting_auto_ptr<SSLClient> ss;
    {
      PythonThreadsAllower all;
      ClientSocket sock(hostname, port, timeout * 1000);
      ss = counting_auto_ptr<SSLClient>(new SSLClient(sock));
      ss->connect(timeout * 1000);
    }
    int id = (int) ss->socket().get_sock();
    ssls[id] = ss;
    return Py_BuildValue("i", id);
  } catch (String e) {
    PyErr_SetString(PyExc_Exception, e.c_str());
  } catch ( ... ) {
    PyErr_SetString(PyExc_Exception, "unknown");
  }
  return NULL;
}

PyObject *
conga_ssl_lib_disconnect(PyObject *self, PyObject *args)
{
  int id;
  if (!PyArg_ParseTuple(args, "i", &id))
    return NULL;

  try {
    map<int, counting_auto_ptr<SSLClient> >::iterator iter =
      ssls.find(id);
    if (iter != ssls.end())
      ssls.erase(iter);
    Py_INCREF(Py_None);
    return Py_None;
  } catch (String e) {
    PyErr_SetString(PyExc_Exception, e.c_str());
  } catch ( ... ) {
    PyErr_SetString(PyExc_Exception, "unknown");
  }
  return NULL;
}

PyObject *
conga_ssl_lib_send(PyObject *self, PyObject *args)
{
  int id;
  const char* msg;
  int timeout;
  if (!PyArg_ParseTuple(args, "isi", &id, &msg, &timeout))
    return NULL;
  if (timeout < 0) {
    PyErr_SetString(PyExc_ValueError, "negative timeout");
    return NULL;
  }

  try {
    map<int, counting_auto_ptr<SSLClient> >::const_iterator iter =
      ssls.find(id);
    if (iter == ssls.end())
      throw String("SSL connection closed");

    {
      PythonThreadsAllower all;
      int beg = int(time_sec());
      String out(msg);
      while (true) {
	if (int(time_sec()) > beg + timeout)
	  throw String("timeout");
	else
	  if ((out = iter->second->send(out, 400)).empty())
	    break;
      }
    }

    Py_INCREF(Py_None);
    return Py_None;
  } catch (String e) {
    PyErr_SetString(PyExc_Exception, e.c_str());
  } catch ( ... ) {
    PyErr_SetString(PyExc_Exception, "unknown");
  }
  return NULL;
}

PyObject *
conga_ssl_lib_recv(PyObject *self, PyObject *args)
{
	int id, timeout;

	if (!PyArg_ParseTuple(args, "ii", &id, &timeout))
		return NULL;

	if (timeout < 0) {
		PyErr_SetString(PyExc_ValueError, "negative timeout");
		return NULL;
	}

	try {
		map<int, counting_auto_ptr<SSLClient> >::const_iterator iter =
			ssls.find(id);

		if (iter == ssls.end())
			throw String("SSL connection closed");

		String resp;
		{
			PythonThreadsAllower all;

			int beg = int(time_sec());
			String xml_in;
			while (true) {
				size_t buflen;
				String ret;
				if (int(time_sec()) > beg + timeout)
					throw String("timeout");
				else {
					ret = iter->second->recv(400, buflen);
					if (ret == "")
						continue;
					xml_in += ret;
				}

				if (ret.size() < buflen)
				{
					resp = xml_in;
					try {
						parseXML(xml_in);
						break;
					} catch (...) {
						continue;
					}
				}
			}
		}

		PyObject* resp_p = Py_BuildValue("s", resp.c_str());
		return resp_p;
	} catch (String e) {
		PyErr_SetString(PyExc_Exception, e.c_str());
	} catch ( ... ) {
		PyErr_SetString(PyExc_Exception, "unknown");
	}

	return NULL;
}

PyObject *
conga_ssl_lib_trust(PyObject *self, PyObject *args)
{
  int id;
  const char* msg;
  if (!PyArg_ParseTuple(args, "is", &id, &msg))
    return NULL;

  try {
    map<int, counting_auto_ptr<SSLClient> >::const_iterator iter =
      ssls.find(id);
    if (iter == ssls.end())
      throw String("SSL connection closed");

    bool resp;
    {
      PythonThreadsAllower all;
      resp = iter->second->trust_peer_cert(msg);
    }

    PyObject* resp_p = Py_BuildValue("i", (resp)?1:0);
    return resp_p;
  } catch (String e) {
    PyErr_SetString(PyExc_Exception, e.c_str());
  } catch ( ... ) {
    PyErr_SetString(PyExc_Exception, "unknown");
  }
  return NULL;
}

PyObject *
conga_ssl_lib_trusted(PyObject *self, PyObject *args)
{
  int id;
  if (!PyArg_ParseTuple(args, "i", &id))
    return NULL;

  try {
    map<int, counting_auto_ptr<SSLClient> >::const_iterator iter =
      ssls.find(id);
    if (iter == ssls.end())
      throw String("SSL connection closed");

    bool resp;
    {
      PythonThreadsAllower all;
      resp = iter->second->peer_cert_trusted();
    }

    PyObject* resp_p = Py_BuildValue("i", (resp)?1:0);
    return resp_p;
  } catch (String e) {
    PyErr_SetString(PyExc_Exception, e.c_str());
  } catch ( ... ) {
    PyErr_SetString(PyExc_Exception, "unknown");
  }
  return NULL;
}

PyObject *
conga_ssl_lib_untrust(PyObject *self, PyObject *args)
{
  int id;
  if (!PyArg_ParseTuple(args, "i", &id))
    return NULL;

  try {
    map<int, counting_auto_ptr<SSLClient> >::const_iterator iter =
      ssls.find(id);
    if (iter == ssls.end())
      throw String("SSL connection closed");

    bool resp;
    {
      PythonThreadsAllower all;
      resp = iter->second->untrust_peer_cert();
    }

    PyObject* resp_p = Py_BuildValue("i", (resp)?1:0);
    return resp_p;
  } catch (String e) {
    PyErr_SetString(PyExc_Exception, e.c_str());
  } catch ( ... ) {
    PyErr_SetString(PyExc_Exception, "unknown");
  }
  return NULL;
}

PyObject *
conga_ssl_lib_peer_fingerprint(PyObject *self, PyObject *args)
{
  int id;
  if (!PyArg_ParseTuple(args, "i", &id))
    return NULL;

  try {
    map<int, counting_auto_ptr<SSLClient> >::const_iterator iter =
      ssls.find(id);
    if (iter == ssls.end())
      throw String("SSL connection closed");

    String finger, digest;
    {
      PythonThreadsAllower all;
      finger = iter->second->peer_cert_fingerprint(digest);
    }

    PyObject* resp_p = Py_BuildValue("(ss)", digest.c_str(), finger.c_str());
    return resp_p;
  } catch (String e) {
    PyErr_SetString(PyExc_Exception, e.c_str());
  } catch ( ... ) {
    PyErr_SetString(PyExc_Exception, "unknown");
  }
  return NULL;
}
